#!/usr/bin/env python3

# Copyright (c) 2020-2021:
#   Helmholtz-Zentrum Potsdam Deutsches GeoForschungsZentrum GFZ
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.


import os
import argparse
import numpy as np
import pandas as pd
import geopandas as gpd
import datetime
import sqlite3
from shapely import wkt

from .fragility_model import FragilityModel
from .losslib import get_full_gm_field


VERSION_STRING = "losscalculator version 0.1"


def damage_calculator(
    exposure_filepath,
    fragility_source,
    intensity_measure_map_filepath,
    taxonomy_map_filepath,
    gm_field_filepath,
    interpolation_method="linear",
    result_filepath="damage_result.csv",
    output_format="gpkg",
    use_xml_fragilities=True,
    aggregation_type="building",
):
    """
    This function computes the probabilities of occurrence (PoO) of damage states for a scenario
    earthquake, given a ground-motion field, an exposure model representing the assets in the
    region of interest and fragility functions that model the probability of exceeding defined
    damage states given a set of ground-motion values of a specific type.
    Please read the manual for more information on the inputs.
    """

    # Initialize variables.
    all_im_values = []
    all_im_types = []
    all_poes = []
    all_poos = []
    damage_by_assets = []

    # Read inputs.
    gm_field = np.loadtxt(gm_field_filepath, delimiter=",", skiprows=1)
    if exposure_format == "csv":
        exposure = pd.read_csv(exposure_filepath)
    else:  # exposure_format == "spatialite"
        try:
            conn = sqlite3.connect(exposure_filepath)
            conn.enable_load_extension(True)
            sql_statement = "SELECT load_extension('%s');" % spatialite_extension_filepath
            conn.execute(sql_statement)
        except sqlite3.OperationalError:
            raise ValueError("Spatialite database cannot be opened. Exiting ...")

        sql_statement = "SELECT BuildingAsset.id, "
        sql_statement += "X(ST_Centroid(Building.geom)) AS lon, "
        sql_statement += "Y(ST_Centroid(Building.geom)) AS lat, "
        sql_statement += "Taxonomy.taxonomy_string AS taxonomy, "
        sql_statement += "BuildingAsset.number, "
        sql_statement += "BuildingAsset.structural, "
        sql_statement += "BuildingAsset.night, "
        sql_statement += "BuildingAsset.occupancy, "
        sql_statement += "District.admin_name, "
        sql_statement += "District.admin_id, "
        sql_statement += "Tile.id AS tile_id, "
        sql_statement += "ST_AsText(Tile.geom) AS tile_geometry, "
        sql_statement += "Building.id AS building_id, "
        sql_statement += "ST_AsText(Building.geom) AS building_geometry "
        sql_statement += "FROM BuildingAsset INNER JOIN Building "
        sql_statement += "ON BuildingAsset.building_id = Building.id "
        sql_statement += "INNER JOIN Tile ON Building.tile_id = Tile.id "
        sql_statement += "INNER JOIN Taxonomy ON BuildingAsset.taxonomy_id = Taxonomy.id "
        sql_statement += "INNER JOIN District ON BuildingAsset.district_id = District.id"
        exposure = pd.read_sql_query(sql_statement, conn)
    fragility_model = FragilityModel(
        fragility_source,
        gm_field_filepath,
        intensity_measure_map_filepath,
        taxonomy_map_filepath,
        use_xml_fragilities,
    )

    # Interpolate the ground-motion values for all the assets of the exposure model.
    # pylint: disable=E1101
    full_gm_field = get_full_gm_field(
        gm_field, exposure.lon, exposure.lat, interpolation_method
    )

    # Going through each asset to do the damage calculation.
    # pylint: disable=E1136
    for asset in range(exposure.shape[0]):
        taxonomy = exposure["taxonomy"][asset]
        num_buildings = exposure["number"][asset]

        # Computing probabilities of exceedance and occurrence.
        poes, poos, im_value, im_type = fragility_model(taxonomy, full_gm_field[asset])
        all_poes.append(poes)
        all_poos.append(poos)
        all_im_values.append(im_value)
        all_im_types.append(im_type)

        # Compute damage by assets.
        damage_by_assets.append([i * num_buildings for i in poos])

    # Append results.
    # pylint: disable=E1137
    exposure["gm_value"] = all_im_values
    exposure["gm_type"] = all_im_types
    exposure["PoEs"] = all_poes
    exposure["PoOs"] = all_poos

    exposure[fragility_model.loss_category + "_no_damage"] = [
        row[0] for row in damage_by_assets
    ]
    limit_state_names = [fragility_model.loss_category + "_no_damage"]

    for i in range(len(fragility_model.limit_states)):
        exposure[fragility_model.loss_category + "_" + fragility_model.limit_states[i]] = [
            row[(i + 1)] for row in damage_by_assets
        ]
        limit_state_names.append(
            fragility_model.loss_category + "_" + fragility_model.limit_states[i]
        )

    if aggregation_type == "building":
        aggregated = exposure.groupby(
            ["id", "building_geometry", "tile_id", "tile_geometry"], as_index=False
        )[["number", "structural", "night"] + limit_state_names].sum()
        if output_format == "gpkg":
            aggregated["geometry"] = aggregated["building_geometry"].apply(wkt.loads)
            aggregated.drop("building_geometry", axis=1, inplace=True)
            aggregated_gdf = gpd.GeoDataFrame(aggregated, geometry="geometry", crs=4326)
            aggregated_gdf.to_file(result_filepath, driver="GPKG")
        else:
            aggregated.to_csv(result_filepath)
    elif aggregation_type == "tile":
        aggregated = exposure.groupby(["tile_id", "tile_geometry"], as_index=False)[
            ["number", "structural", "night"] + limit_state_names
        ].sum()
        if output_format == "gpkg":
            aggregated["geometry"] = aggregated["tile_geometry"].apply(wkt.loads)
            aggregated.drop("tile_geometry", axis=1, inplace=True)
            aggregated_gdf = gpd.GeoDataFrame(aggregated, geometry="geometry", crs=4326)
            aggregated_gdf.to_file(result_filepath, driver="GPKG")
        else:
            aggregated.to_csv(result_filepath)
    else:
        exposure.to_csv(result_filepath, index=False)


def main():

    parser = argparse.ArgumentParser(
        description="This program computes the probabilities of occurrence of different "
        + "damage states for a scenario earthquake, given a ground-motion field "
        + "for the area, an exposure model representing the assets in the region "
        + "of interest and fragility functions that show the probability of "
        + "exceeding a set of damage states, given an intensity measure level."
    )
    parser.add_argument(
        "command",
        type=str,
        help="'damage' to run a damage assessment, "
        + "'version' to output the version information.",
    )
    parser.add_argument(
        "-i",
        "--interpolation-method",
        required=False,
        type=str,
        default="linear",
        choices=["linear", "cubic", "nearest"],
        help="method used to interpolate the input ground-motion values over all "
        + "the exposure locations (asset locations). "
        + "Options: ['linear', 'nearest', 'cubic'] (default set to linear)",
    )
    parser.add_argument(
        "-f",
        "--fragility-file",
        required=False,
        type=str,
        help="file path to a fragility-function XML file",
    )
    parser.add_argument(
        "-c",
        "--csv-directory",
        required=False,
        type=str,
        help="path name of the directory containing fragility functions as CSV files (*.csv)",
    )
    parser.add_argument(
        "-m",
        "--intensity-measure-map",
        required=False,
        type=str,
        help="file path to the file that includes the mapping from the intensity-measure "
        + "names in the ground-motion field to the intensity-measure name used in the "
        + "fragility functions. If not provided, a default set will be used",
    )
    parser.add_argument(
        "-g",
        "--ground-motion-field",
        required=False,
        type=str,
        help="path to the ground-motion values file. This file should include all the "
        + "intensity measure types that you have in your fragility functions",
    )
    parser.add_argument(
        "-e",
        "--exposure",
        required=False,
        type=str,
        help="path to the file that includes the exposure assets. Either a CSV file "
        + "or a Spatialite database.",
    )
    parser.add_argument(
        "-r",
        "--results",
        required=False,
        type=str,
        default="damage_result.gpkg",
        help="path to the file to write the results to (default: damage_result.gpkg) ",
    )
    parser.add_argument(
        "-o",
        "--output-format",
        required=False,
        type=str,
        default="gpkg",
        choices=["gpkg", "csv"],
        help="format of the result output, either Geopackage ('gpkg') or comma-separated "
        + "values file ('csv'). Output to Geopackage only works without option '-a'."
        + "Options: ['gpkg', 'csv'] (default: 'gpkg')",
    )
    parser.add_argument(
        "-w",
        "--overwrite",
        required=False,
        action="store_true",
        help="overwrite result file if it already exists",
    )
    parser.add_argument(
        "-a",
        "--aggregation",
        required=False,
        type=str,
        default="building",
        choices=["no", "building", "tile"],
        help="type of result aggregation. 'no' provides detailed results for debugging "
        + "and does not work with output to Geopackage files ('-o gpkg).",
    )
    parser.add_argument(
        "-t",
        "--taxonomy-map",
        required=False,
        type=str,
        help="file path to the file that includes the mapping from complex taxonomy "
        + "strings to simple taxonomy strings as provided by the fragility "
        + "functions. If not provided, taxonomies will be passed directly to "
        + "the fragility functions.",
    )
    args = parser.parse_args()

    # read arguments from command line.
    command = args.command
    interpolation_method = args.interpolation_method
    fragility_source = args.fragility_file
    csv_directory = args.csv_directory
    intensity_measure_map = args.intensity_measure_map
    gm_field_filepath = args.ground_motion_field
    exposure_filepath = args.exposure
    result_filepath = args.results
    output_format = args.output_format
    overwrite_result_file = args.overwrite
    aggregation_type = args.aggregation
    taxonomy_map_filepath = args.taxonomy_map

    if command == "version":
        print(VERSION_STRING)
    elif command == "damage":
        # Check for necessary command line options
        if gm_field_filepath is None:
            print(
                "Error: No ground-motion field provided. Please use the "
                + "-g/--ground-motion-field option to provide a ground-motion field file."
            )
            exit()
        if exposure_filepath is None:
            print(
                "Error: No exposure data provided. Please use the "
                + "-e/--exposure option to provide an exposure file."
            )
            exit()
        use_xml_fragilities = True
        if fragility_source is not None:
            if csv_directory is not None:
                print("-f option given, ignoring -c option.")
        else:
            if csv_directory is not None:
                fragility_source = csv_directory
                use_xml_fragilities = False
            else:
                print(
                    "Error: No fragility functions provided. Please use the "
                    + "-f/--fragility-file option to provide a fragility-function file or "
                    + "-c/--csv-directory option to provide a directory containing "
                    + "fragility functions as CSV files."
                )
                exit()
        if output_format == "gpkg" and aggregation_type == "no":
            print("Error: Cannot output to Geopackage file without aggregation (-a option). ")
            exit()
        if os.path.exists(result_filepath):
            if not overwrite_result_file:
                print(
                    "Error: The result file exists. Please choose another file name or set "
                    + "--overwrite to overwrite the existing result file."
                )
                exit()
            else:
                os.remove(result_filepath)

        startTime = datetime.datetime.now()
        print(startTime)

        damage_calculator(
            exposure_filepath,
            fragility_source,
            intensity_measure_map,
            taxonomy_map_filepath,
            gm_field_filepath,
            interpolation_method,
            result_filepath,
            output_format,
            use_xml_fragilities,
            aggregation_type,
        )
        print("Execution time of the script", (datetime.datetime.now() - startTime))
    else:
        print("Command unknown. Exiting.")


if __name__ == "__main__":
    main()
