import csv
import numpy as np
import xml.etree.ElementTree as ETree
import glob
from scipy.stats import lognorm
from pathlib import Path


class FragilityModel:
    """
    General class to handle a complete fragility model as a set of fragility functions
    for different structural types
    """

    def __init__(
        self,
        fragility_source,
        gmf_filepath,
        intensity_measure_map_filepath=None,
        taxonomy_map_filepath=None,
        use_xml_fragilities=True,
    ):
        """
        Initializing the fragility model by importing either an XML file with fragility
        functions or a set of fragility functions stored in CSV files from a directory
        and creating the intensity-measure map that maps intensity-measure types as given in
        the fragility functions to the ones in the ground-motion field file. It then creates
        a mapping dictionary between the intensity measure types and their respective
        column number in the ground-motion field file for faster access to data.
        This function is limited to one fragility function per building taxonomy.

        Dictionary structure for discrete functions (XML or CSV):
        'CR/LFINF+CDL/HBET:7-20/0.0':
            {'imv': array([2.225000e-03, 2.397000e-03, 2.581000e-03]),
             'slight': array([0.00000e+00, 0.00000e+00, 0.00000e+00]),
             'moderate': array([0.00000e+00, 0.00000e+00, 0.00000e+00]),
             'extensive': array([0.00000e+00, 0.00000e+00, 0.00000e+00]),
             'complete': array([0.00000e+00, 0.00000e+00, 0.00000e+00]),
             'format': 'discrete',
             'imt': 'SA(1)'}

        Dictionary structure for a continuous function:
        'CR/LFINF+CDL/HBET:7-20/0.0':
            {'slight': {'mean': "1.692", 'stddev': "1.179"},
             'moderate': {'mean': "1.692", 'stddev': "1.179"},
             'extensive': {'mean': "1.692", 'stddev': "1.179"},
             'complete': {'mean': "1.692", 'stddev': "1.179"},
             'format': 'continuous',
             'imt': 'SA(1)',
             'maxIML': "5",
             'minIML': "1e-10",
             'noDamageLimit': "0.05",
             'function': function object}

        Mapping dictionary example:
        {'PGA': 2, 'SA(0.3)': 3, 'SA(0.6)': 4, 'SA(1)': 5}

        Args:
            fragility_source (str):
                File path of an XML file containing fragility functions in NRML as
                provided by the Global Earthquake Model or a file search pattern
                for CSV files containing fragility functions (e.g.: 'csvfiles/*.csv')

            gmf_filepath (str):
                File path to the ground-motion field.
                (Header line contains the intensity-measure names)

                File example extract:

                lon,lat,gmv_PGA,gmv_SA(0.3),gmv_SA(0.6),gmv_SA(1.0)
                22.89028,36.31806,4.822530E-03,1.255729E-02,8.736364E-03,5.612176E-03
                22.89028,36.32083,4.830462E-03,1.257595E-02,8.748890E-03,5.619673E-03
                ...

            intensity_measure_map_filepath (str):
                File path to a CSV file mapping the intensity types from the
                ground-motion field to intensity types as given in the fragility functions.
                If `None`, a default mapping will be applied.

                File example extract:

                in_fragility_function, in_ground_motion_field
                SA(0.3),gmv_SA(0.3)
                SA(0.6),gmv_SA(0.6)
                SA(1),gmv_SA(1.0)
                SA(1.0),gmv_SA(1.0)
                PGA,gmv_PGA

            taxonomy_map_filepath (str):
                File path to the mapping of taxonomy strings to their relevant
                fragility function names.

                File example extract:

                taxonomy,conversion
                CR+PC/LFM+CDN/H:1,CR_LFM-CDN_H1_0
                CR/LFM+CDN/H:1,CR_LFM-CDN_H1_0
                ...

            use_xml_fragilities (boolean)
                If set (default), an XML file is expected in fragility_source, otherwise
                the file search pattern for CSV files.

        """

        self.fragility_functions = {}
        self.gm_type_column_map = {}
        self.taxonomy_map = {}
        self.limit_states = None
        self.num_limit_states = 0

        # Use XML file
        if use_xml_fragilities:
            tree = ETree.parse(fragility_source)
            root = tree.getroot()[0]
            assert "fragilityModel" in root.tag, (
                "%s does not contain a recognized fragility model" % fragility_source
            )

            self.asset_category = root.attrib["assetCategory"]
            self.loss_category = root.attrib["lossCategory"]
            for child in root:
                if "description" in child.tag:
                    self.description = child.text
                elif "limitStates" in child.tag:
                    self.limit_states = child.text.split()
                    self.num_limit_states = len(self.limit_states)
                elif "fragilityFunction" in child.tag:
                    limit_states = []
                    # Intensity-measure type
                    im_type = None
                    parameters = {}
                    # Import a discrete function
                    if child.get("format") == "discrete":
                        # Intensity-measure values and probabilities of exceedance
                        for ichild in list(child):
                            key, value = ichild.items()[0]
                            if key == "imt":
                                im_type = value
                                im_value = np.array(list(map(float, ichild.text.split())))
                                parameters["imv"] = im_value
                            if key == "ls":
                                ls = value
                                values = np.array(list(map(float, ichild.text.split())))
                                parameters[ls] = values
                                limit_states.append(value)
                        parameters["format"] = child.get("format")
                        parameters["imt"] = im_type
                        if im_type is None:
                            raise ValueError(
                                "No intensity measure given for fragility function %s"
                                % (child.get("id"))
                            )
                        if not (limit_states == self.limit_states):
                            raise ValueError(
                                "Limit states for fragility function"
                                + "%s differ from global limit states %s"
                                % (child.get("id"), self.limit_states)
                            )

                    # Import a continuous function
                    elif child.get("format") == "continuous":
                        parameters["format"] = child.get("format")
                        if child.get("shape") == "logncdf":
                            for ichild in list(child):
                                if "imls" in ichild.tag:
                                    for key, value in ichild.items():
                                        parameters[key] = value
                                        if key == "imt":
                                            im_type = value
                                elif "params" in ichild.tag:
                                    ls_params = {}
                                    limit_state = None
                                    for key, value in ichild.items():
                                        if key == "ls":
                                            limit_state = value
                                            limit_states.append(value)
                                        else:
                                            ls_params[key] = float(value)
                                    variance = ls_params["stddev"] ** 2.0
                                    sigma = np.sqrt(
                                        np.log((variance / (ls_params["mean"] ** 2.0)) + 1.0)
                                    )
                                    mu = (ls_params["mean"] ** 2.0) / np.sqrt(
                                        variance + ls_params["mean"] ** 2.0
                                    )
                                    ls_params["function"] = lognorm(sigma, scale=mu)

                                    parameters[limit_state] = ls_params
                        else:
                            raise ValueError(
                                "Unknown function form %s of fragility function %s"
                                % (child.get("shape"), child.get("id"))
                            )

                    # Final checks to make sure the fragility function is valid
                    if im_type is None:
                        raise ValueError(
                            "No intensity measure given for fragility function %s"
                            % (child.get("id"))
                        )
                    if not (limit_states == self.limit_states):
                        raise ValueError(
                            "Limit states for fragility function"
                            + "%s differ from global limit states %s"
                            % (child.get("id"), self.limit_states)
                        )
                    self.fragility_functions[child.get("id")] = parameters

        # Use CSV files
        else:
            self.asset_category = "buildings"
            self.loss_category = "structural"
            self.description = ""

            for ff_filepath in glob.glob(fragility_source):
                limit_states = []
                parameters = {"format": "discrete"}
                taxonomy = Path(ff_filepath).stem
                with open(ff_filepath) as f:
                    line = f.readline()
                    line_list = line.split(",")
                    parameters["imt"] = line_list[0]
                    values = np.array(list(map(float, line_list[1:])))
                    parameters["imv"] = values
                    for line in f:
                        line_list = line.split(",")
                        limit_states.append(line_list[0])
                        values = np.array(list(map(float, line_list[1:])))
                        parameters[line_list[0]] = values

                # Final checks to make sure the fragility function is valid
                if self.limit_states is None:
                    self.limit_states = limit_states
                    self.num_limit_states = len(self.limit_states)
                else:
                    if not (limit_states == self.limit_states):
                        raise ValueError(
                            "Limit states for fragility function"
                            + "%s differ from global limit states %s" % taxonomy,
                            self.limit_states,
                        )
                self.fragility_functions[taxonomy] = parameters

        # Read the intensity measure map into a dictionary
        if intensity_measure_map_filepath is not None:
            with open(intensity_measure_map_filepath, mode="r") as input_file:
                reader = csv.reader(input_file)
                next(reader)
                self.intensity_measure_map = {rows[0]: rows[1] for rows in reader}
        else:
            self.intensity_measure_map = {
                "SA(0.3)": "gmv_SA(0.3)",
                "SA(0.6)": "gmv_SA(0.6)",
                "SA(1.0)": "gmv_SA(1.0)",
                "SA(1)": "gmv_SA(1.0)",
                "PGA": "gmv_PGA",
                "MMI": "gmv_MMI",
            }

        # Read the taxonomy map if filename provided
        if taxonomy_map_filepath is not None:
            with open(taxonomy_map_filepath, mode="r") as input_file:
                reader = csv.reader(input_file)
                next(reader)
                self.taxonomy_map = {rows[0]: rows[1] for rows in reader}

        # Extracting the ground-motion types available in the ground-motion field.
        with open(gmf_filepath) as gmf:
            im_types = gmf.readline().strip().split(",")[2:]
        im_types_dict = {}
        for index, value in enumerate(im_types):
            im_types_dict[value] = index + 2
        # Add for each intensity measure from the fragility functions the respective
        # column number to the dictionary. Hereby, using the intensity-measure map to
        # translate between the name in the fragility functions and the name in
        # ground-motion field.
        for im_ff in self.intensity_measure_map:
            column = im_types_dict.get(self.intensity_measure_map[im_ff], None)
            if column is not None:
                self.gm_type_column_map[im_ff] = column

    def __repr__(self):
        # Informative string representation
        return "Fragility Model: %s\nLimit States: %s" % (
            self.description,
            " ".join(self.limit_states),
        )

    def __len__(self):
        # Return the number of fragility functions in the model
        return len(self.fragility_functions)

    def __call__(self, taxonomy, intensity_measure_values):
        """
        Computing PoE values for all the different damage states available in the
        fragility functions, considering the ground-motion value.

        Args:
            taxonomy (str):
                Taxonomy string of an asset
                Example:
                >>> taxonomy
                'CR/LDUAL+CDM/HBET:6-/11.0'
            intensity_measure_values (list):
                Full row of the ground-motion field for one asset represented as a list
                containing the lognitude, latitude, and all avaialble intensity measure
                values.
                Example:
                >>> intensity_measure_values
                [22.89028,36.31806,4.822530E-03,1.255729E-02,8.736364E-03,5.612176E-03]

        Returns:
            poes (list):
                A list with element as probabilities of exceedance.
                Example:
                >>> poes
                [0.73, 0.52, 0.32, 0.18]
            poos (list):
                A list with element as probabilities of occurence.
                Example:
                >>> poos
                [0.27, 0.21, 0.2 , 0.14, 0.18]
            im_value (float):
                Intensity-measure value used to compute the PoEs.
            im_type (str):
                Intensity-measure type used to compute the PoEs.
        """

        # Apply the taxonomy mapping
        tax = self.taxonomy_map.get(taxonomy, taxonomy)
        # Retrieve the fragility function matching the taxonomy
        ff = self.fragility_functions[tax]
        # Retrieve the necessary intensity-measure type and value for the given taxonomy
        im_type = ff["imt"]
        column = self.gm_type_column_map.get(im_type, None)
        if column is None:
            raise ValueError(
                "Necessary intensity measure type %s not given in ground-motion file." % im_type
            )
        im_value = intensity_measure_values[column]

        if ff["format"] == "discrete":
            if im_value <= ff["imv"][0]:
                # The intensity-measure value is smaller than or equal to the minimum
                # intensity-measure value of the fragility function
                poes = [1.0]
                poes.extend([0.0] * self.num_limit_states)
                poos = list(poes)
            elif im_value >= ff["imv"][-1]:
                # The intensity-measure value is larger than or equal to the maximum
                # intensity-measure value of the fragility function
                poes = [1.0]
                for ls in self.limit_states:
                    poes.append(ff[ls][-1])
                poos = list(-1 * np.diff(poes))
                poos.append(poes[-1])
            else:
                # The intensity-measure value is in the range of defined
                # intensity-measure values of the fragility function
                # Determining  iml that is smaller than the given ground-motion value
                iml_min = ff["imv"][ff["imv"] < im_value].max()
                # Finding index of the iml_min
                index_iml_min = np.searchsorted(ff["imv"], iml_min, side="left")
                # Smallest iml that is larger than the given ground-motion value
                iml_max = ff["imv"][ff["imv"] > im_value].min()
                # Finding index of the iml_max
                index_iml_max = np.searchsorted(ff["imv"], iml_max, side="left")
                # Two bounds of the interpolation
                x = [iml_min, iml_max]
                # Prepare return value
                poes = [1.0]
                # Looping over all limit states
                for ls in self.limit_states:
                    y = [ff[ls][index_iml_min], ff[ls][index_iml_max]]
                    # Interpolating PoE values for the given gm-value between x and y
                    # range (using interpolation)
                    poes.append(np.interp(im_value, x, y))

        # Process a continuous fragility function
        else:
            # If the minimum, maximum and no-damage limit intensity measures are defined
            # then retrieve them and clip in input ground motion intensity measure
            min_iml = float(ff.get("minIML", 0.0))
            max_iml = float(ff.get("maxIML", np.inf))
            # no_damage_limit = float(ff.get("noDamageLimit", 0.0))
            if im_value < min_iml:
                im_value = min_iml
            if im_value > max_iml:
                im_value = max_iml
            # Prepare return value
            poes = [1.0]
            # Get the PoEs
            for ls in self.limit_states:
                # Interpolating PoE values for the given gm-value between x and y
                # range (using interpolation)
                poes.append(ff[ls]["function"].cdf(im_value))

        # Calculating PoOs
        # pylint: disable=E1130
        poos = list(-1 * np.diff(poes))
        poos.append(poes[-1])
        return poes[1:5], poos, im_value, im_type
