#!/usr/bin/env python3

# Copyright (c) 2020-2021:
#   Helmholtz-Zentrum Potsdam Deutsches GeoForschungsZentrum GFZ
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.


import numpy as np
from scipy.interpolate import griddata


def get_full_gm_field(gm_field, lons, lats, method="linear"):

    """
    Returns ground-motion values using 2-dimensional interpolation for the
    given longitudes and latitudes. The deafult interpolation method is linear.

    Input:
    ------
    - gm_field: (Numpy nd-array of shape(m,n); m records with long,
        lat, and n-2 gm-types)
        contains locations and the ground-motion values of each location
        in the format of lon, lat, gmValueofType1, gmValueofType2, ...,
        gmValueofTypeN.
        Please note that if the gm_field contains values for
        more than one ground-motion type, This function will do
        interpolation multiple times and each time using one of the
        gmValueTypes with the format as below:
        [lon, lat, gmValueofType1, gmValueofType2, ..., gmValueofTypeN]
        As an example:
        [23.6875,38.3402777778,'PGA','SA(0.3)','SA(0.6)','SA(1.0)']
        Example extract:
        >>> gm_field
        array([
            [2.289e+01, 3.631e+01, 4.822e-03, 1.255e-02, 8.736e-03, 5.612e-03],
            [2.289e+01, 3.632e+01, 4.830e-03, 1.257e-02, 8.748e-03, 5.619e-03],
            ...,
            [2.410e+01, 3.818e+01, 4.961e-02, 1.078e-01, 7.710e-02, 4.852e-02]
            ])

    - lons: (pandas.DataFrame(series) OR numpy-1darray of len a)
        longitudes of the locations we want to interpolate over.
        Example extract:
        >>> lons
        23.6875
        ...
        24.0264

    - lats: (pandas.DataFrame(series) OR numpy-1darray of len a)
        latitudes of the locations we want to interpolate over.
        Example extract:
        >>> lats
        38.340278
        ...
        37.651389

    - method: (string), optional
        {‘linear’, ‘nearest’, ‘cubic’}, optional'. Linear by default.

    Output:
    -------
    - full_gm_field: (numpy-ndarray of shape(a,n))
        an array of logitudes and latitudes along with their interpolated values
        Example extract:
        >>> full_gm_field
        array([
        [2.36e+01, 3.83e+01, 1.79e-01, 3.62e-01, 2.52e-01, 1.52e-01],
        ...,
        [2.40e+01, 3.76e+01, 2.03e-02, 4.61e-02, 3.07e-02, 1.81e-02]
        ])
    """

    # points_given: the input points for interpolation
    points_given = np.vstack((gm_field[:, 0], gm_field[:, 1])).transpose()
    # points_todo: points to do interpolation over
    points_todo = np.vstack((lons, lats)).transpose()
    full_gm_field = np.vstack((np.array(lons), np.array(lats))).transpose()
    # Griddata Interpolation
    # The loop changes over the columns of the ground motion field and enables
    # us to have interpolation for all different given ground-motion types.
    # Please note that the columns follow the order below:
    # lon,lat,gmValueofType1,gmValueofType2,etc. We want to do the interpolation
    # over all the ground motion types. Thus the range begins from third column
    # to the last column of the gm_field.
    for gm_type in range(2, gm_field.shape[1], 1):
        # gmvs_given: the input ground-motion values for interpolation.
        gmvs_given = gm_field[:, gm_type]
        # gm_value_griddata : interpolated values for each ground motion type.
        gm_value_griddata = griddata(points_given, gmvs_given, points_todo, method=method)
        full_gm_field = np.column_stack((full_gm_field, gm_value_griddata))
    return full_gm_field
