#!/usr/bin/env python3

# Copyright (C) 2020-2021:
#   Helmholtz-Zentrum Potsdam Deutsches GeoForschungsZentrum GFZ
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

import logging
import csv
import sqlite3
import glob
import numpy as np
import shapely.geometry as shgeometry
from shapely.geometry import box
from pygeotile.tile import Tile

# Initialize log
logger = logging.getLogger(__name__)


def get_coordinates_from_cell_id(cell_id):
    """
    Given a cell ID, it returns the coordinates that define the four corners of the cell.
    """
    if 8398080000 > cell_id > -1:
        lon_edges = np.linspace(-180.0, 180.0, num=129601)
        lat_edges_inv = np.linspace(-90.0, 90.0, num=64801)
        cell_y = -(cell_id // 129600) + 64799
        cell_x = cell_id - 8397950400 + 129600 * cell_y
        lon_w = lon_edges[cell_x]
        lon_e = lon_edges[cell_x + 1]
        lat_s = lat_edges_inv[cell_y]
        lat_n = lat_edges_inv[cell_y + 1]
    else:
        print("ERROR!! CELL ID " + str(cell_id) + " DOES NOT EXIST!!!")
        lon_w = -999.9
        lon_e = -999.9
        lat_s = -999.9
        lat_n = -999.9
    return lon_w, lat_s, lon_e, lat_n


def get_geometry_wkt_of_cell(cell_id):
    lon_w, lat_s, lon_e, lat_n = get_coordinates_from_cell_id(cell_id)
    out_wkt = (
        shgeometry.Polygon([(lon_w, lat_s), (lon_e, lat_s), (lon_e, lat_n), (lon_w, lat_n)])
    ).wkt
    return out_wkt


class Database:
    """The Database class represents a Spatialite database. It manages the database
    connection and cursor. Upon connection, the Spatialite extension is loaded.

    Args:
        database_filepath (str):
            File path for the Spatialite database file.
        spatialite_filepath (str):
            File path of the Spatialite extension.

    Attributes:
        database_filepath (str): Spatialite database file path.
        spatialite_filepath (str): File path to the Spatialite extension.
    """

    def __init__(self, database_filepath, spatialite_filepath="mod_spatialite"):
        self.database_filepath = database_filepath
        self.spatialite_filepath = spatialite_filepath

        self.connection = None
        self.cursor = None

    def create_connection_and_cursor(self, init_spatial_metadata=True):
        """Create a database connection, loads the Spatialite extension and initializes it."""

        # Create SQLite database for the exposure data
        logger.debug("Connecting to database at %s" % self.database_filepath)
        self.connection = sqlite3.connect(self.database_filepath)
        logger.debug("Connection to database established")

        self.connection.enable_load_extension(True)
        sql_statement = "SELECT load_extension('%s');" % self.spatialite_filepath
        self.connection.execute(sql_statement)
        if init_spatial_metadata:
            self.connection.execute("SELECT InitSpatialMetaData(1);")
            logger.debug("Spatialite extension loaded and initialized")

        versions = self.connection.execute("SELECT sqlite_version(), spatialite_version()")
        for row in versions:
            logger.debug("SQLite version: %s" % row[0])
            logger.debug("Spatialite version: %s" % row[1])

        self.cursor = self.connection.cursor()


class ExposureDatabase(Database):
    """
    The ExposureDatabase class represents a Spatialite database for
    exposure models. It is derived from the generic Database class.

    Args:
        database_filepath (str):
            File path for the Spatialite database file.
        spatialite_filepath (str):
            File path of the Spatialite extension.

    Attributes:
        database_filepath (str): Spatialite database file path.
        spatialite_filepath (str): File path to the Spatialite extension.
    """

    def __init__(self, database_filepath, spatialite_filepath="mod_spatialite"):
        Database.__init__(self, database_filepath, spatialite_filepath)

    def create_tables(self):
        """
        Creates all necessary tables in the database. These are:
            BuildingAsset: Stores building-type assets
            TileAsset    : Stores tile-type assets
            Building     : Stores building geometries
            Tile         : Stores tile geometries
            Taxonomy     : Stores taxonomy strings
            District     : Stores administrative districts
        """

        # Create table BuildingAsset
        sql_statement = "CREATE TABLE BuildingAsset ("
        sql_statement += "id           TEXT, "
        sql_statement += "taxonomy_id  INTEGER, "
        sql_statement += "number       REAL, "
        sql_statement += "structural   REAL, "
        sql_statement += "night        REAL, "
        sql_statement += "occupancy    TEXT, "
        sql_statement += "district_id  TEXT, "
        sql_statement += "building_id  INTEGER)"
        self.connection.execute(sql_statement)
        logger.debug("Table BuildingAsset created")

        # Create table TileAsset
        sql_statement = "CREATE TABLE TileAsset ("
        sql_statement += "id           TEXT, "
        sql_statement += "taxonomy_id  INTEGER, "
        sql_statement += "number       REAL, "
        sql_statement += "structural   REAL, "
        sql_statement += "night        REAL, "
        sql_statement += "occupancy    TEXT, "
        sql_statement += "district_id  TEXT, "
        sql_statement += "tile_id      INTEGER)"
        self.connection.execute(sql_statement)
        logger.debug("Table TileAsset created")

        # Create table Buildings
        sql_statement = "CREATE TABLE Building ("
        sql_statement += "id       INTEGER, "
        sql_statement += "tile_id  INTEGER)"
        self.connection.execute(sql_statement)
        sql_statement = "SELECT AddGeometryColumn('Building', 'geom', 4326, "
        sql_statement += "'MULTIPOLYGON', 'XY')"
        self.connection.execute(sql_statement)
        logger.debug("Table Building created")

        # Create table Tiles
        sql_statement = "CREATE TABLE Tile ("
        sql_statement += "id  INTEGER)"
        self.connection.execute(sql_statement)
        sql_statement = "SELECT AddGeometryColumn('Tile', 'geom', 4326, 'POLYGON', 'XY')"
        self.connection.execute(sql_statement)
        logger.debug("Table Tile created")

        # Create table Taxonomy
        sql_statement = "CREATE TABLE Taxonomy ("
        sql_statement += "id               INTEGER, "
        sql_statement += "taxonomy_string  TEXT)"
        self.connection.execute(sql_statement)
        logger.debug("Table Taxonomy created")

        # Create table District
        sql_statement = "CREATE TABLE District ("
        sql_statement += "id          INTEGER, "
        sql_statement += "admin_id    TEXT, "
        sql_statement += "admin_name  TEXT)"
        self.connection.execute(sql_statement)
        logger.debug("Table District created")

    def insert_building_asset(
        self,
        building_asset_id,
        taxonomy_id,
        number,
        structural,
        night,
        occupancy,
        district_id,
        building_id,
    ):
        """
        Inserts a building-type asset to the BuildingAssets table.

        Args:
            building_asset_id (str):
                ID of the building asset
            taxonomy_id (int):
                ID if the taxonomy of the tile asset (corresponds to Taxonomy.id)
            number (float):
                Number of buildings of the tile asset
            structural (float):
                Structural value of the tile asset (sum over all buildings in this asset)
            night (float):
                Number of people in the buildings of the tile asset (sum)
            occupancy (str):
                Occupancy type of the tile asset
            district_id (int):
                ID of the district (corresponds to District.id)
            building_id (int):
                ID of the building(corresponds to Building.id)
        """

        sql_statement = "INSERT INTO BuildingAsset "
        sql_statement += "(id, taxonomy_id, number, structural, night, "
        sql_statement += "occupancy, district_id, building_id) "
        sql_statement += "VALUES ('%s', %d, %f, %f, %f, '%s', %d, %d)" % (
            building_asset_id,
            taxonomy_id,
            number,
            structural,
            night,
            occupancy,
            district_id,
            building_id,
        )
        self.cursor.execute(sql_statement)

    def insert_tile_asset(
        self,
        tile_asset_id,
        taxonomy_id,
        number,
        structural,
        night,
        occupancy,
        district_id,
        tile_id,
    ):
        """
        Inserts a tile-type asset to the TileAssets table.

        Args:
            tile_asset_id (str):
                ID of the tile asset
            taxonomy_id (int):
                ID if the taxonomy of the tile asset (corresponds to Taxonomy.id)
            number (float):
                Number of buildings of the tile asset
            structural (float):
                Structural value of the tile asset (sum over all buildings in this asset)
            night (float):
                Number of people in the buildings of the tile asset (sum)
            occupancy (str):
                Occupancy type of the tile asset
            district_id (int):
                ID of the district (corresponds to District.id)
            tile_id (int):
                ID of the tile the tile asset is located in
        """

        sql_statement = "INSERT INTO TileAsset "
        sql_statement += "(id, taxonomy_id, number, structural, night, "
        sql_statement += "occupancy, district_id, tile_id) "
        sql_statement += "VALUES ('%s', %d, %f, %f, %f, '%s', %d, %d)" % (
            tile_asset_id,
            taxonomy_id,
            number,
            structural,
            night,
            occupancy,
            district_id,
            tile_id,
        )
        self.cursor.execute(sql_statement)

    def insert_building(self, building_id, geom, tile_id):
        """
        Inserts a building and its geometry to the Building table.

        Args:
            building_id (int):
                ID of the building
            geom (geom):
                Geometry of the building (multi-polygon)
            tile_id (int):
                ID of the tile the building centroid is located in
        """

        sql_statement = "INSERT INTO Building "
        sql_statement += "(id, geom, tile_id) "
        sql_statement += "VALUES (%d, %s, %d)" % (building_id, geom, tile_id)
        self.cursor.execute(sql_statement)

    def insert_tile(self, tile_id, use_tiles=True):
        """
        Inserts a tile ID and its respective geometry to the Tile table.
        The geometry is computed from the tile ID before inserting.

        Args:
            tile_id (int):
                ID of the tile
            use_tiles (boolean):
                Use Quadtree tile IDs instead of IDs of 10-arcsecond cells
        """

        if use_tiles:
            tile = Tile.from_quad_tree(str(tile_id))
            tile_polygon = box(
                tile.bounds[0].longitude,
                tile.bounds[0].latitude,
                tile.bounds[1].longitude,
                tile.bounds[1].latitude,
            )
            tile_geometry = tile_polygon.wkt
        else:
            tile_geometry = get_geometry_wkt_of_cell(tile_id)
        sql_statement = "INSERT INTO Tile "
        sql_statement += "(id, geom) "
        sql_statement += "VALUES (%d, %s)" % (
            tile_id,
            "GeomFromText('%s', 4326)" % tile_geometry,
        )
        self.cursor.execute(sql_statement)

    def insert_taxonomy(self, taxonomy_id, taxonomy_string):
        """
        Inserts a taxonomy ID and its respective description to the Taxonomy table.

        Args:
            taxonomy_id (int):
                ID of the taxonomy item
            taxonomy_string (str):
                Taxonomy string following the GEM Taxonomy
        """

        sql_statement = "INSERT INTO Taxonomy "
        sql_statement += "(id, taxonomy_string) "
        sql_statement += "VALUES (%d, '%s')" % (taxonomy_id, taxonomy_string)
        self.cursor.execute(sql_statement)

    def insert_district(self, district_id, admin_id, admin_name):
        """
        Inserts a district ID and name to the District table.

        Args:
            district_id (int):
                ID of the district dataset
            admin_id (str):
                Official ID of the district
            admin_name (str):
                Official Name of the district
        """

        sql_statement = "INSERT INTO District "
        sql_statement += "(id, admin_id, admin_name) "
        sql_statement += "VALUES (%d, '%s', '%s')" % (district_id, admin_id, admin_name)
        self.cursor.execute(sql_statement)

    def import_exposure(
        self, import_search_pattern, building_geometries_filepath, use_tiles=True
    ):
        """
        Import an exposure model given as OpenQuake-type exposure CSV files to the database.
        In case the model contains assets describing buildings, the building polygons
        need to be provided in an extra CSV file.

        Args:
            import_search_pattern (str):
                The pattern of the exposure model files to import (e.g. 'data/*.csv').
            building_geometries_filepath (str):
                File path of the building geometries (in EPSG:3857) as CSV file. The file
                includes the building IDs (labeled `origin-id`), their corresponding geometries
                and IDs of the tiles in which the building centroids are located.
                See the example test for a sample version of this file.
            use_tiles (boolean):
                Use Quadtree tile IDs instead of IDs of 10-arcsecond cells
        """

        self.create_tables()
        tile_ids = []
        if building_geometries_filepath:
            with open(building_geometries_filepath) as csv_file:
                reader = csv.reader(csv_file, delimiter=";")
                next(reader)
                for building in reader:
                    building_id = int(building[0][4:])
                    # Removes the first 5 characters (i.e. the part 'cell_').
                    tile_id = int(building[2][5:])
                    if tile_id not in tile_ids:
                        tile_ids.append(tile_id)
                    self.insert_building(
                        building_id,
                        # If the geometry is given in 3857, it needs to be transformed into 4326
                        # "ST_Transform(CastToMultiPolygon(GeomFromText('%s', 3857)), 4326)"
                        # % building[1],
                        "GeomFromText('%s', 4326)" % building[1],
                        tile_id,
                    )
                logger.info("Buildings imported")
                self.connection.commit()
        else:
            print(
                "Please provide the input `-g`,`--building-geometries` if there are any "
                + "building assets in the input CSV format exposure file."
            )
        # Create temporary storage for taxonomy strings and districts
        taxonomy_list = []
        district_list = []

        # Open input exposure file
        file_list = glob.glob(import_search_pattern)
        for exposure_filepath in file_list:
            logger.info("Processing %s" % exposure_filepath)
            with open(exposure_filepath) as csv_file:
                reader = csv.reader(csv_file, delimiter=",")
                next(reader)
                line_number = 0
                for asset in reader:

                    taxonomy = asset[3]
                    if taxonomy not in taxonomy_list:
                        taxonomy_list.append(taxonomy)
                    taxonomy_index = taxonomy_list.index(taxonomy)

                    district = [asset[8], asset[9]]
                    if district not in district_list:
                        district_list.append(district)
                    district_id = district_list.index(district)

                    if asset[10].startswith("OSM_"):
                        building_id = int(asset[10][4:])
                        self.insert_building_asset(
                            asset[0],
                            taxonomy_index,
                            float(asset[4]),
                            float(asset[5]),
                            float(asset[6]),
                            asset[7],
                            district_id,
                            building_id,
                        )
                    else:
                        tile_id = int(asset[10][5:])
                        self.insert_tile_asset(
                            asset[0],
                            taxonomy_index,
                            float(asset[4]),
                            float(asset[5]),
                            float(asset[6]),
                            asset[7],
                            district_id,
                            tile_id,
                        )
                        if tile_id not in tile_ids:
                            tile_ids.append(tile_id)

                    line_number += 1
                    if line_number % 100000 == 0:
                        logger.info("Imported asset %d" % line_number)
                logger.info("Assets imported")
                self.connection.commit()

        for taxonomy in taxonomy_list:
            self.insert_taxonomy(taxonomy_list.index(taxonomy), taxonomy)
        self.connection.commit()
        logger.debug("Taxonomies stored")

        for district in district_list:
            self.insert_district(district_list.index(district), district[0], district[1])
        self.connection.commit()
        logger.debug("Districts stored")

        for tile_id in tile_ids:
            self.insert_tile(tile_id, use_tiles=use_tiles)
        self.connection.commit()
        logger.debug("Tiles generated and stored")

    def export_exposure(self, export_filepath):
        """
        Exports an exposure CSV file.

        Export file example extract:
        "id","lon","lat","taxonomy","number","structural","night","occupancy",
            "admin_name","admin_id","tile_id","tile_geometry","building_id",
            "building_geometry"
        "GDE_Ind_64",23.71122313388688,37.97458194956127,"CR/LFINF+CDM/H:2",
            0.225532,402838.427402,0.0,"Ind","GR_3514501","Athens",2427351736,
            "POLYGON((23.711111 37.972222, 23.711111 37.972222))",458630414,
            "MULTIPOLYGON(((23.711173 37.974563, 23.711236 37.974539, 23.71117 37.974563)))"

        Args:
            export_filepath (str):
                File path to the augmented exposure model CSV file.
        """

        export_file = open(export_filepath, mode="w")
        fieldnames = [
            "id",
            "lon",
            "lat",
            "taxonomy",
            "number",
            "structural",
            "night",
            "occupancy",
            "admin_name",
            "admin_id",
            "tile_id",
            "tile_geometry",
            "building_id",
            "building_geometry",
        ]
        writer = csv.writer(
            export_file,
            delimiter=",",
            quoting=csv.QUOTE_NONNUMERIC,
        )
        writer.writerow(fieldnames)

        # Building asset exposure output
        sql_statement = "SELECT BuildingAsset.id, "
        sql_statement += "X(ST_Centroid(Building.geom)), "
        sql_statement += "Y(ST_Centroid(Building.geom)), "
        sql_statement += "Taxonomy.taxonomy_string, "
        sql_statement += "BuildingAsset.number, "
        sql_statement += "BuildingAsset.structural, "
        sql_statement += "BuildingAsset.night, "
        sql_statement += "BuildingAsset.occupancy, "
        sql_statement += "District.admin_name, "
        sql_statement += "District.admin_id, "
        sql_statement += "Tile.id, "
        sql_statement += "ST_AsText(Tile.geom), "
        sql_statement += "Building.id, "
        sql_statement += "ST_AsText(Building.geom) "
        sql_statement += "FROM BuildingAsset INNER JOIN Building "
        sql_statement += "ON BuildingAsset.building_id = Building.id "
        sql_statement += "INNER JOIN Tile ON Building.tile_id = Tile.id "
        sql_statement += "INNER JOIN Taxonomy ON BuildingAsset.taxonomy_id = Taxonomy.id "
        sql_statement += "INNER JOIN District ON BuildingAsset.district_id = District.id"
        self.cursor.execute(sql_statement)
        logger.debug("Output statement executed")

        line_number = 0
        for asset in self.cursor:
            write_asset = [
                asset[0],
                asset[1],
                asset[2],
                asset[3],
                asset[4],
                asset[5],
                asset[6],
                asset[7],
                asset[8],
                asset[9],
                asset[10],
                asset[11],
                asset[12],
                asset[13],
            ]
            writer.writerow(write_asset)
            line_number += 1
            if line_number % 10000 == 0:
                logger.info("Exported line " + str(line_number))
        logger.info("Building exposure exported")

        # Tile asset exposure output
        sql_statement = "SELECT TileAsset.id, "
        sql_statement += "X(ST_Centroid(Tile.geom)), "
        sql_statement += "Y(ST_Centroid(Tile.geom)), "
        sql_statement += "Taxonomy.taxonomy_string, "
        sql_statement += "TileAsset.number, "
        sql_statement += "TileAsset.structural, "
        sql_statement += "TileAsset.night, "
        sql_statement += "TileAsset.occupancy, "
        sql_statement += "District.admin_name, "
        sql_statement += "District.admin_id, "
        sql_statement += "Tile.id, "
        sql_statement += "ST_AsText(Tile.geom) "
        sql_statement += "FROM TileAsset "
        sql_statement += "INNER JOIN Tile ON TileAsset.tile_id = Tile.id "
        sql_statement += "INNER JOIN Taxonomy ON TileAsset.taxonomy_id = Taxonomy.id "
        sql_statement += "INNER JOIN District ON TileAsset.district_id = District.id"
        self.cursor.execute(sql_statement)
        logger.debug("Output statement executed")

        for asset in self.cursor:
            write_asset = [
                asset[0],
                asset[1],
                asset[2],
                asset[3],
                asset[4],
                asset[5],
                asset[6],
                asset[7],
                asset[8],
                asset[9],
                asset[10],
                asset[11],
                "-1",
                "POINT EMPTY",
            ]
            writer.writerow(write_asset)
            line_number += 1
            if line_number % 10000 == 0:
                logger.info("Exported line " + str(line_number))
        export_file.close()
        logger.info("Tile exposure exported")
