#!/usr/bin/env python3

# Copyright (C) 2020-2021:
#   Helmholtz-Zentrum Potsdam Deutsches GeoForschungsZentrum GFZ
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

"""
Returns an exposure model along with the asset geometries as either a CSV or a database file.
Please use either `import` to import exposure data into the database or `export` to export to
an OpenQuake-compatible exposure CSV file.

Program Inputs:
---------------
`-d`, `--exposure-database`: str, required
Defines the path to an exposure model database file.
When using import module, it points to a file where the exposure database file is to be stored.
When using export module, it points the an existing exposure database file from which a CSV
exposure file with header
`id,lon,lat,taxonomy,number,structural,night,occupancy,admin_name,admin_ID, tile_id,
tile_geometry, building_id ,building_geometry` is exported.

`-g`, `--building-geometries`: str, optional
Defines the path to the file that maps the building ID (`origin-id` for a building asset
in the exposure model) from which the asset is originated to the building geometry and the tile
ID that the building centroid is located in.
Only required if there are building assets in the input exposure model; with format as below:
`origin_id;geometry;tile_id`

`-i`, `--import-search-pattern`: str, optional
Search pattern for the input exposure model files (e.g. `*.csv`).
It should follow the format below:
`id,lon,lat,taxonomy,number,structural,night,occupancy,admin_name,admin_ID,origin_id`
Only required when using the import module.

`-o`, `--export-filepath`: str, optional
Defines the filepath to the output CSV exposure model file (OpenQuake compatible) with format
as below:
`id,lon,lat,taxonomy,number,structural,night,occupancy,admin_name,admin_ID, tile_id,
tile_geometry, building_id ,building_geometry`
Only required when using the export module.

`-s`, `--spatialite-extension`: str, optional
Defines the file path of the spatialite extension. (default: mod_spatialite)

`-w` overwrite an existing result file: optional
To overwrite an existing exposure model file.

In order to use this program, please note:
The spacing of the grid is 10 arc-seconds. The tile ID starts from the north-west corner of the
world, moves east per row, and finishes at the south-east corner of the world. First tile ID
is 0, last tile ID is 8,398,079,999 (total number of tiles is 8,398,080,000).


Execution examples:
---------------
Create a database exposure model file:
python3 preprocessor.py import -d ../example_preprocessor/augmented_exposure.db
-i ../example_preprocessor/data/exposure_*.csv
-g ../example_preprocessor/data/building_to_cellid_map.csv

Create a CSV exposure model file:
python3 preprocessor.py export -d ../example_preprocessor/augmented_exposure.db
-o ../example_preprocessor/augmented_exposure.csv

"""

import logging
import sys
import os
import argparse
import sqlite3
from database import ExposureDatabase

# Add a logger printing error, warning, info and debug messages to the screen
logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
logger.addHandler(logging.StreamHandler(sys.stdout))


def main():
    """
    Returns an exposure model along with the asset geometries as either a CSV or database file.
    """
    parser = argparse.ArgumentParser(
        description="This program imports exposure data into an exposure database "
        + "(Spatialite) from OpenQuake-style exposure data and OBM building data."
    )
    help_str = "'import' to import exposure data into the database, "
    help_str += "'export' to export to an OpenQuake-compatible exposure CSV file"
    parser.add_argument(
        "command",
        type=str,
        help=help_str,
    )
    parser.add_argument(
        "-d",
        "--exposure-database",
        required=True,
        type=str,
        help="File path to the exposure model database file (Required)",
    )
    parser.add_argument(
        "-i",
        "--import-search-pattern",
        required=False,
        type=str,
        help="Search pattern for the input exposure model files for importing (e.g. '*.csv')",
    )
    parser.add_argument(
        "-o",
        "--export-filepath",
        required=False,
        type=str,
        help="File path to the exposure model CSV file",
    )
    parser.add_argument(
        "-g",
        "--building-geometries",
        required=False,
        type=str,
        help="File path of the CSV file with building geometries (in EPSG:3857) "
        + "following the format: `building_id;geometry;cell_id`",
    )
    parser.add_argument(
        "-w",
        "--overwrite",
        required=False,
        action="store_true",
        help="Overwrite the existing exposure database file",
    )
    parser.add_argument(
        "-s",
        "--spatialite-extension",
        required=False,
        type=str,
        default="mod_spatialite",
        help="File path of the spatialite extension (default: mod_spatialite)",
    )
    parser.add_argument(
        "-c",
        "--cells",
        required=False,
        action="store_true",
        help="Use 10-arcsecond cells instead of Quadtree tiles",
    )

    args = parser.parse_args()

    command = args.command
    exposure_database = args.exposure_database
    export_filepath = args.export_filepath
    import_search_pattern = args.import_search_pattern
    building_geometries_filepath = args.building_geometries
    overwrite_exposure_database = args.overwrite
    spatialite_filepath = args.spatialite_extension
    use_tiles = not args.cells

    if command == "import":
        if os.path.exists(exposure_database):
            if overwrite_exposure_database:
                os.remove(exposure_database)
            else:
                logger.warning("Database exists. Exiting ...")
                exit()
        db = ExposureDatabase(exposure_database, spatialite_filepath)
        try:
            db.create_connection_and_cursor()
        except sqlite3.OperationalError:
            logger.warning("Spatialite extension cannot be loaded. Exiting ...")
            exit()
        db.import_exposure(
            import_search_pattern, building_geometries_filepath, use_tiles=use_tiles
        )
    elif command == "export":
        db = ExposureDatabase(exposure_database, spatialite_filepath)
        try:
            db.create_connection_and_cursor(init_spatial_metadata=False)
        except sqlite3.OperationalError:
            logger.warning("Spatialite extension cannot be loaded. Exiting ...")
            exit()
        db.export_exposure(export_filepath)


main()
